# frozen_string_literal: true

require 'bundler/setup'
require 'semver_dialects'
require 'simplecov'
require 'rspec-parameterized'

SimpleCov.start

RSpec.configure do |config|
  # Enable flags like --only-failures and --next-failure
  config.example_status_persistence_file_path = '.rspec_status'

  # Disable RSpec exposing methods globally on `Module` and `main`
  config.disable_monkey_patching!

  config.expect_with :rspec do |c|
    c.syntax = :expect
  end
end

RSpec.shared_examples 'table-driven test' do |input, want, _got|
  context "given input #{input}" do
    it "returns #{want}" do
      expect(subject).to eq(want)
    end
  end
end
